/*
  Mètode de Newton-Raphson en el pla complex.

  Autor: Daniel Fernández Benítez.
  Última modificació: 19/06/2021.
  
  
  L'usuari proporciona una funció complexa a què aplicar el mètode de Newton-
Raphson, junt amb una condició inicial i un valor per al criteri d'aturada
del mètode iteratiu.
  Com a criteri d'aturada del mètode iteratiu escollim que la distància
entre dos iterats consecutius sigui inferior a un valor prefixat proporcionat
per l'usuari. També s'imposa un nombre màxim d'iteracions.


-------------------------------
  Les dades que l'usuari pot tenir interès a modificar (funcio, condició inicial,
constants, etc.) es troben en MÈTODE DE NEWTON.
-------------------------------


POSSIBLES MILLORES:
  - Rebre la condició inicial i la distància entre iterats com arguments
   del programa.
  - Afegir el mètode de Newton amortiguat i permetre que l'usuari en triï
   quin fer servir amb un argument del programa.
*/



#include<stdio.h>
#include<stdlib.h>
#include<complex.h>
#include<math.h>
#include<stdbool.h>


/*-----------------------------------------------------------
                      MÈTODE DE NEWTON
-----------------------------------------------------------*/

/*Informació relativa a l'execució del mètode de Newton:
condició inicial, condicions d'aturada, la funció.*/


/*Nombre màxim d'iteracions*/
#define MAX_ITER 15
/*criteri d'aturada: distància entre iterats consecutius*/
#define DIST_ITERS_CONSEC 1E-10

/*tolerància, al quadrat, per considerar un nombre com 0 (TOL = 1E-12)*/
#define TOL2_NULLITAT 1E-24


/*condició inicial per al mètode de Newton*/
#define NEWTON_COND_INI (-1. + 1. * I)

/*funció a què aplicar el mètode de Newton*/
#define funcio(z) (z * z * z - 1.)
//#define funcio(z) (-3 + z * (7 + z * z * (-2 + z)))

/*derivada de la funció a què aplicar el mètode de Newton*/
#define funcio_1(z) (3. * z * z)
//#define funcio_1(z) (7 + z * z * (-6 + 4 * z))
/*-----------------------------------------------------------
                    FI MÈTODE DE NEWTON
-----------------------------------------------------------*/


/*mòdul al quadrat d'un complex (cabs(z) ^ 2)*/
#define cmodul2(z) (creal(z) * creal(z) + cimag(z) * cimag(z))



char * g_nom_prog;


int newton_raphson(double complex *, const double);




int main(int argc, char *argv[]){
  int q_iter;
  double complex z_0, z_sol;
  
  
  /*processar arguments del programa*/
  g_nom_prog = argv[0];


  /*iniciar el mètode iteratiu de Newton*/
  z_sol = z_0 = NEWTON_COND_INI;
  q_iter = newton_raphson(&z_sol, DIST_ITERS_CONSEC);
  
  
  /*informem sobre el resultat*/
  if(z_sol == z_0){
    printf("Hi ha hagut algun problema amb el procés d'iteració," \
        " en la iteració número %d\n", q_iter);
  }else{
    printf("iteracions: %d\n" \
        "z_0: %16.9le %+16.9le i\n" \
        "z_sol: %16.9le %+16.9le i\n", q_iter, creal(z_0), cimag(z_0),
        creal(z_sol), cimag(z_sol));
  }
  
   
    
  return EXIT_SUCCESS;
}


/*
DESC: Mètode de Newton en el pla complex.
  Executa el mètode iteratiu de Newton amb la condició inicial donada, El mètode
s'atura amb èxit si s'ha satisfet el criteri d'aturada: la distància
entre dos iterats consecutius és inferior a «dist_iters_atur»

              |x_{k+1} - x_k| < «dist_iters_atur».

  En tal cas es retorna l'últim iterat calculat, desant-lo al paràmetre «z».
En qualsevol altre cas d'aturada, «z» no es modifica: la derivada s'anula
en un iterat o s'arriba al màxim d'iteracions.

PARAMS:
  - *z, apuntador a la condició inicial del mètode de Newton. Es desarà
   l'últim iterat calculat en cas d'aturar-se amb èxit.
  - dist_iters_atur, la distància entre dos iterats consecutius per a aturar
   el

RETORN: El nombre d'iteracions fetes fins que el mètode s'ha aturat.
*/
int newton_raphson(double complex *z, const double dist_iters_atur){
  const double dist2_iters_atur = dist_iters_atur * dist_iters_atur;
  int q_iter;
  double complex z0, z1, f_1;
  bool es_divisio_zero;
  
  
  if(z == NULL){
    return 0;
  }
  
  
  /*mètode de Newton-Raphson*/
  q_iter = 0;
  z0 = z1 = *z;
  es_divisio_zero = false;
  do{
    z0 = z1;
    
    /*abans d'iterar, comprovem si la derivada s'hi anul·la*/
    f_1 = funcio_1(z0);
    if(cmodul2(f_1) < TOL2_NULLITAT){
      es_divisio_zero = true;
      continue;
    }
    
    /*iteració Newton*/
    z1 = z0 - funcio(z0) / f_1;

    
    q_iter++;
  }while(! es_divisio_zero && q_iter < MAX_ITER && 
      cmodul2(z1 - z0) > dist2_iters_atur);
  
  /*desem la solució a «z»*/
  if(! es_divisio_zero && cmodul2(z1 - z0) <= dist2_iters_atur){
    *z = z1;
  }
  
  
  return q_iter;
}

