/*
  Programa per a calcular els punts en què se satisfan
les hipòtesis del Teorema de Kantorovich per a una funció complexa donada.
  El programa escriu els punts a la sortida estàndar; per tant cal redirigir
la sortida per a desar-ho a un arxiu i poder fer-ne imatges amb el Gnuplot.

      ./conques-atraccio-complexos.out > arxiu_escriure_punts.dad


  Autor: Daniel Fernández Benítez.
  Última modificació: 18/06/2021.

  
  L'usuari proporciona una funció complexa i una bola oberta en el pla complex
a on la segona derivada de la funció està afitada.
  Aleshores es crea una xarxa de punts en un rectangle que conté la bola,
i en cada punt de la xarxa que sigui dins la bola es comproven les hipòtesis
del Teorema de Kantorovich. Tots els punts en què se satisfan les hipòtesis
de Kantorovich s'escriuen a la sortida estàndar.


-------------------------------
  Les dades que l'usuari pot tenir interès a modificar (funcio, constants, etc.)
es troben en XARXA_PUNTS i TEOREMA KANTOROVICH.
-------------------------------


  El programa s'ha escrit definint com a macros del preprocessador aquells
càlculs que s'han de repetir molts cops i que, altrament, caldria escriure
en una funció. D'aquesta manera s'evita que el programa estigui constantment
fent crides a funcions, i així accelerem l'execució del programa.



POSSIBLES MILLORES
 - Potser fer els gràfics amb Gnuplot directament des del programa
  amb el mòdul http://ndevilla.free.fr/gnuplot/
*/



#include<stdio.h>
#include<stdlib.h>
#include<complex.h>
#include<math.h>
#include<stdbool.h>


/*-----------------------------------------------------------
                     TEOREMA KANTOROVICH
-----------------------------------------------------------*/
/*tolerància per considerar un nombre com 0*/
#define TOL_NULLITAT 1E-12


/*funció a què aplicar el mètode de Newton*/
#define funcio(z) (z * z * z - 1.)
//#define funcio(z) (-3 + z * (7 + z * z * (-2 + z)))

/*derivada de la funció a què aplicar el mètode de Newton*/
#define funcio_1(z) (3 * z * z)
//#define funcio_1(z) (7 + z * z * (-6 + 4 * z))


/*Bola que conté els zeros de la funció definida dalt (f(z))*/
#define BOLA_CENTRE (0. + 0. * I)
#define BOLA_RADI 2.
//#define BOLA_RADI 2.5


/*M = fita de |f''(z)| en la bola*/
#define KANTOR_M 12.
//#define KANTOR_M 45.

/*Condició Kantorovich h = (beta)*M*(eta) <= 1/2*/
#define MAX_H_KANTOR 1. / 2
/*-----------------------------------------------------------
                    FI TEOREMA KANTOROVICH
-----------------------------------------------------------*/


/*-----------------------------------------------------------
                        XARXA PUNTS
-----------------------------------------------------------*/
/*Domini rectangular en què crear la xarxa de punts.
Ha de contenir la bola definida dalt*/
#define X_MIN -2.
#define Y_MAX 2.
#define X_MAX 2.
#define Y_MIN -2.

/*quantitat de punts en què dividir cada costat del rectangle*/
#define Q_PUNTS_XARXA_X 512
#define Q_PUNTS_XARXA_Y Q_PUNTS_XARXA_X

/*distància (horitzontal i vertical) entre punts de la xarxa*/
#define XARXA_PAS_X  ((X_MAX - X_MIN) / (Q_PUNTS_XARXA_X - 1))
#define XARXA_PAS_Y  ((Y_MAX - Y_MIN) / (Q_PUNTS_XARXA_Y - 1))
/*-----------------------------------------------------------
                      FI XARXA PUNTS
-----------------------------------------------------------*/


/*mòdul al quadrat d'un complex (cabs(z) ^ 2)*/
#define cmodul2(z) (creal(z) * creal(z) + cimag(z) * cimag(z))

/*Comprova que un complex pertany a la bola (tancada) definida dalt*/
#define pertany_bola(z) \
  (cmodul2(z - BOLA_CENTRE) + TOL_NULLITAT <= BOLA_RADI * BOLA_RADI)
  
/*comprova que una bola (tancada) B(z, r)
està continguda en la bola (oberta) definida dalt*/
#define es_bola_dins(z, r) \
  (cmodul2(z - BOLA_CENTRE) + TOL_NULLITAT < (BOLA_RADI - r) * (BOLA_RADI - r))



char * g_nom_prog;


void conques_Kantorovich();




int main(int argc, char *argv[]){
  /*processar arguments del programa*/
  g_nom_prog = argv[0];

  
  /*calcular i escriure els punts en què es compleix la condició de Kantorovich*/
  conques_Kantorovich();
   
    
  return EXIT_SUCCESS;
}


/*
DESC: Comprova en quins punts dins la bola es compleixen les hipòtesis
del Teorema de Kantorovich i els escriu a la sortida estàndar.
  Per a cada punt de la xarxa es comprova el següent
    - si el punt és dins la bola.
    - si el valor h = (beta) * M * (eta) és menor que 1/2.
    - si la regió d'existència, que dona el Teorema per al punt, és dins la bola.

si compleix tot això, compleix les hipòtesis del Teorema de Kantorovich
i escrivim el punt a la sortida estàndar.

PARAMS:

RETORN:
*/
void conques_Kantorovich(){
  unsigned int i, j;
  double complex z;
  double beta, eta, h, r_exist;
  const double dx = XARXA_PAS_X, dy = XARXA_PAS_Y;
  
  
  /*comprovem les hipòtesis del Tª Kantorovich
  en els punts de la xarxa que pertanyen a la bola*/
  for(i = 0; i < Q_PUNTS_XARXA_Y; i++){
    for(j = 0; j < Q_PUNTS_XARXA_X; j++){
      z = (X_MIN + j * dx) + (Y_MAX - i * dy) * I;
      
      /*comprovem si pertany a la bola*/
      if(! pertany_bola(z)){
        continue;
      }
      
      
      /*càlcul de h = (beta) * M * (eta)  */
      /*beta = |1 / f'(z)| */
      beta = cabs(funcio_1(z));
      if(beta < TOL_NULLITAT){
        continue;
      }
      beta = 1. / beta;
      
      /*eta = |f(z) / f'(z)|*/
      eta = cabs(funcio(z)) * beta;
      h = beta * KANTOR_M * eta;
      
      
      /*si es compleixen les hipòtesis del Tª Kantorovich, escrivim el punt*/
      /*r = (eta) * (1 - sqrt(1 - 2h)) / h) =
          = 2h / (beta * M * (1 + sqrt(1 - 2h)))*/
      r_exist = 2. * h / (beta * KANTOR_M * (1 + sqrt(1 - 2 * h)));
      if(h + TOL_NULLITAT <= MAX_H_KANTOR && es_bola_dins(z, r_exist)){
        printf("%13.7le %13.7le\n", creal(z), cimag(z));
      }
    }
  }
}
